//***************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                         *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                 *//
//* http://apps.jcns.fz-juelich.de/frida                                    *//
//***************************************************************************//

//! \file  opr.cpp
//! \brief NOperate, NIntOld: operate on data files

#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include <boost/format.hpp>

#include <read-plus/ask.hpp>
#include <read-plus/readln.hpp>

#include "defs.h"
#include "olf.h"
#include "mem.h"
#include "zentry.h"
#include "reg.h"
#include "var.h"
#include "expr.h"
#include "opr.h"
#include "xax_lex.h"

using boost::format;

//***************************************************************************//
//* operate via generic expressions                                         *//
//***************************************************************************//

//! Display data. Use boolean expression to display only a subset of the data.

void NOperate::Show( const string& subcmd ) 
{
    NOlm::IterateD fiter;

    static string expr;
    expr = sask( "Show if", expr);
    if (expr=="") expr = "1";
    PNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw "dummy argument t not allowed when operating on data";

    while( POld fin=fiter() ) {
        uint k = fiter.k();
        if ( fiter.size()>1 )
            printf( ".. file %d:\n", fiter.k() );

        double lval;
        string outj, outp;

        // determine length of longest spectrum:
        uint ni = 0;
        for ( uint j=0; j<fin->nJ(); j++)
            ni = max( ni, fin->VS(j)->size() );

        // printout by i:
        for ( uint i=0; i<ni; i++) {

            outj = "";
            outp = "";

            for( uint j=0; j<fin->nJ(); j++ ){
                const PSpec s = fin->VS(j);
                if( i>= s->size() ) continue;
                T->tree_point_val( &lval, 0, k, j, i );
                if( !lval ) continue;
                outj += S(j) + " ";
                if      ( subcmd=="xyd" )
                    if( s->dy.size() )
                        outp += str( format( "%- 12.5g %- 12.5g +- %- 12.5g" )
                                     % s->x[i] % s->y[i] % s->dy[i] );
                    else
                        outp += str( format( "%- 12.5g %- 12.5g" )
                                     % s->x[i] % s->y[i] );
                else if ( subcmd=="y" )
                    outp += str( format( "%- 12.5g" ) % s->y[i] );
                else
                    throw "invalid subcmd " + subcmd;
                outp += " ";
            }

            if( outp=="" ) continue;

            printf( "%5u [%s] %s\n", i, outj.substr(0,outj.size()-1).c_str(), 
                    outp.substr(0,outp.size()-1).c_str() );
        }
    }
}


//! Delete or retain data points according to a boolean expression.

void NOperate::Select( bool sel_del )
{
    NOlm::IterateO fiter;

    static string expr;
    expr = sask((sel_del ? "Delete" : "Retain"), expr);
    if (expr=="") return;
    PNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw "dummy argument t not allowed when operating on data";

    while ( POlo fin=fiter() ) {
        POld fd = P2D( fin );
        POlc fc = P2C( fin );

        POlo fout( fin->new_olo() );
        fout->lDoc.push_back((sel_del ? "md " : "mr ") + expr);

        for ( uint j=0; j<fin->nJ(); j++ ) {
            if ( fd ) {
                PSpec ein = fd->VS(j);
                PSpec eout = PSpec( new CSpec );
                eout->copy_z_base( ein );
                vector<double> lvec( ein->size() );
                T->tree_vec_val( &lvec, 0, fiter.k(), j );
                bool with_dy = ein->dy.size();
                for ( uint i=0; i<ein->size(); i++ ) {
                    if ( sel_del ^ (bool) lvec[i] ) {
                        if( with_dy )
                            eout->push_xyd( ein->x[i], ein->y[i], ein->dy[i] );
                        else
                            eout->push_xy( ein->x[i], ein->y[i] );
                    }
                }
                if ( eout->size() )
                    fout->V.push_back( eout );
            } else if ( fc ) {
                PCurve eout = PCurve( new CCurve( *(fc->VC(j)) ) );
                double lval;
                T->tree_point_val( &lval, 0, fiter.k(), j );
                if ( sel_del ^ (bool) lval )
                    fout->V.push_back( eout );
            }                
        }
        if( !fout->nJ() )
            throw "no data point left";
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Pointwise operation on x, y, or z<n>, according to arithmetic expression.

void NOperate::Pointwise( string llabel ) 
{
    NOlm::IterateO fiter;

    CVar lref(llabel);

    string expr = sask("Expression for " + llabel + " ?");
    if (expr=="") return;
    PNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw "dummy argument t not allowed when operating on data";
 
    while( POlo fin=fiter() ){
        uint k = fiter.k();
        POld fd = P2D( fin );
        POlc fc = P2C( fin );

        if ( lref.pointwise() && !fd )
            throw "no pointwise operation on curve";
        if( lref.typ==CVar::_CP && !fc )
            continue;

        POlo fout( fin->new_olo() );

        if( lref.typ!=CVar::_CP )
            fout->lDoc.push_back( "o" + lref.var_info() + " " + expr);
        
        CCoord co;
        T->set_coord( co, k );
        if        ( lref.typ==CVar::_X ) {
            fout->xco = co;
        } else if ( lref.typ==CVar::_Y ) {
            fout->yco = co;
        } else if ( lref.typ==CVar::_Z ) {
            if( lref.num==-1 )
                fout->ZCo.push_back( co );
            else if( lref.num>= fin->ZCo.size() )
                throw "no such z coordinate to operate on";
            else
                fout->ZCo[lref.num] = co;
        } else if ( lref.typ==CVar::_R ) {
            if( lref.num==-1 )
                fout->RPar.push_back( CParam( co ) );
            else if( lref.num>= fin->RPar.size() )
                throw "no such real par to operate on";
            else
                fout->RPar[lref.num].Co = co;
        } else if ( lref.typ==CVar::_CP ) {
            if( !fc )
                continue; // just ignore op operations on data files
            if( lref.num>=fc->nP )
                throw "no such parameter to operate on";
            NOlm::overwrite = 1;
        }

        if ( lref.typ==CVar::_R ) {
            fout->copy_mainvec( fin );
            double val;
            T->tree_point_val( &val, 0, k, 0 );
            if( lref.num==-1 )
                fout->RPar.back().val = val;
            else
                fout->RPar[lref.num].val = val;
        } else {
            fout->V.resize( fin->nJ() );
            for( uint j=0; j<fin->nJ(); j++ ){
                fout->V[j] = fin->new_zentry( j );
                if ( fin->V[j]->protect )
                    continue;
                if ( lref.pointwise() ) {
                    PSpec eout( new CSpec( *(fd->VS(j)) ) );
                    vector<double> vout( fd->nPts(j) );
                    if ( lref.typ==CVar::_X || lref.typ==CVar::_DY ) {
                        T->tree_vec_val( &vout, 0, k, j );
                        if      ( lref.typ==CVar::_X )
                            (P2D(fout))->VS(j)->x = vout;
                        else if ( lref.typ==CVar::_DY )
                            (P2D(fout))->VS(j)->dy = vout;
                        else
                            throw "BUG: unexpected lref";
                    } else if ( lref.typ==CVar::_Y ) {
                        vector<double> dvout;
                        T->tree_vec_val( &vout, &dvout, k, j );
                        (P2D(fout))->VS(j)->y = vout;
                        if( dvout.size() )
                            (P2D(fout))->VS(j)->dy = dvout;
                    } else
                        throw "BUG: unexpected lref";
                } else if ( lref.typ==CVar::_Z ) {
                    double val;
                    T->tree_point_val( &val, 0, k, j );
                    if ( lref.num==-1 )
                        fout->V[j]->z.push_back( val );
                    else
                        fout->V[j]->z[lref.num] = val;
                } else if ( lref.typ==CVar::_CP ) {
                    double val;
                    T->tree_point_val( &val, 0, k, j );
                    (P2C(fout))->VC(j)->P[lref.num] = val;
                } else {
                    throw "BUG: invalid oxy";
                }
            }
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Integral operation, using arithmetic expression.

void NOperate::Integral(void) 
{
    NOlm::IterateO fiter;

    string expr = sask( "Functional ?" );
    if ( expr=="" ) return;
    PNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw "dummy argument t not allowed when operating on data";

    double res;
    while( POlo fin=fiter() ) {
        uint k = fiter.k();
        uint nz = fin->ZCo.size();
        bool savable = nz > 0;
        
        POlo fout;
        PSpec sout;
        if ( savable ) {
            fout = POlo( fin->new_olo() );
            fout->lDoc.push_back( "oi " +expr );
            fout->ZCo.pop_back();
            fout->xco = fin->ZCo.back();
            T->set_coord( fout->yco, k );
            sout = PSpec( new CSpec );
            sout->z = fin->V[0]->z;
            sout->z.pop_back();
        }

        for ( uint j=0; j<fin->nJ(); j++ ) {
            double xval, zval, dres;
            T->tree_point_val( &res, &dres, k, j );
            if (!savable) {
                cout << "result: " << res;
                if ( dres )
                    cout << " +- " << dres;
                cout << "\n";
                continue;
            }
            xval = fin->V[j]->z[nz-1];
            sout->push_xyd( xval, res, dres );
            if (nz>=2) { // new spectrum if jump in other z values
                zval = fin->V[j]->z[nz-2];
                if ( j+1 < fin->nJ() && fin->V[j+1]->z[nz-2] != zval ) {
                    fout->V.push_back( sout );
                    sout = PSpec( new CSpec );
                    sout->z = fin->V[j+1]->z;
                    sout->z.pop_back();
                }
            }
        }
        if ( savable ) {
            sout->purge_dy();
            fout->V.push_back( sout );
            NOlm::mem_store( fout, fiter.k() );
        }
        if ( !savable )
            NReg::lastresult = res;
    }
};


//! Integral operations to generate not only new y, but also new x.

void NOperate::IntXY(void) 
{
    NOlm::IterateD fiter;

    static int icolx = 0, icoly = 1;
    static CCoord xco, yco;
    icolx = iask( "x from column", icolx );
    if( icolx<0) return;
    icoly = iask( "y from column", icoly );
    if( icoly<0) return;
    xco = CCoord( "x", "" );
    yco = CCoord( "y", "" );

    while( POld fin=fiter() ) {
        uint nz = fin->ZCo.size();

        POld fout( fin->new_old() );
        fout->lDoc.push_back( "oixy " + S(icolx) + " " + S(icoly) );

        fout->xco = xco;
        fout->yco = yco;

        PSpec sout( new CSpec );

        for ( uint j=0; j<fin->nJ(); j++ ) {
            uint n = fin->nPts(j);
            if( icolx>=n || icoly>=n )
                throw "not enough columns in spectrum " + S(j);
            sout->push_xy( fin->VS(j)->y[icolx], fin->VS(j)->y[icoly] );
            sout->z = fin->VS(j)->z;
            if ( nz ) { // new spectrum if jump in any z values
                if ( j+1 < fin->nJ() && fin->V[j+1]->z != fin->V[j]->z ) {
                    fout->V.push_back( sout );
                    sout = PSpec( new CSpec );
                }
            }
        }
        fout->V.push_back( sout );
        NOlm::mem_store( fout, fiter.k() );
    }
};


//***************************************************************************//
//* old-style integral operations                                           *//
//***************************************************************************//

namespace NIntOld {
    static int mod;
    double evaluate( const PSpec& S );
    static double arg2;
    static int iarg2;
};

void NIntOld::Opr() 
{
    NOlm::IterateD fiter;

    if( NRead::stack_empty() )
        cout <<
            "integral properties:\n"
            "   1:                     2: one y\n"
            "   3: <y> average         4: stddev in y\n"
            "   5: x(max y)            6: y(max y)\n"
            "   7: x(min y)            8: y(min y)\n"
            "   9: sum                10: integral\n"
            "  11: x(int<arg)\n"
            "  13: max x(y>0)\n"
            "  15: center of gravity  16: stddev in x\n"
            "  17: correl.coeff (x<->y)\n";
    mod = iask("Choice", mod );
	
    if        (mod== 0) {
        return;
    } else if (mod== 2) {
        iarg2 = iask("channel number", iarg2);
    } else if (mod== 3) {
    } else if (mod== 4) {
    } else if (mod== 5) {
    } else if (mod== 6) {
    } else if (mod== 7) {
    } else if (mod== 8) {
    } else if (mod== 9) {
    } else if (mod==10) {
    } else if (mod==11) {
        arg2 = dask("second argument", arg2);
    } else if (mod==13) {
    } else if (mod==15) {
    } else if (mod==16) {
    } else if (mod==17) {
    } else
        throw "invalid option";
	
    // CList FOpr = NOlm::FSel;
    while( POld fin=fiter() ) {
        int nz = fin->ZCo.size();
        bool savable = nz > 0;

        POld fout( fin->new_old() );

        if( savable ) {
            fout->ZCo.pop_back();
            fout->xco = fin->ZCo.back();

            if        (mod== 2) {
                fout->yco.name = fin->yco.name + " (#"
                    + S(iarg2) + ")";
            } else if (mod== 3) {
                fout->yco.name = "<" + fin->yco.name + ">";
                fout->yco.unit = fin->yco.unit;
            } else if (mod== 4) {
                fout->yco.name = "sigma(" + fin->yco.name + ")";
                fout->yco.unit = "" ;
            } else if (mod== 5) {
                fout->yco.name = fin->xco.name + "_max";
                fout->yco.unit = fin->xco.unit;
            } else if (mod== 6) {
                fout->yco.name = fin->yco.name + "_max";
                fout->yco.unit = fin->yco.unit;
            } else if (mod== 7) {
                fout->yco.name = fin->xco.name + "_min";
                fout->yco.unit = fin->xco.unit;
            } else if (mod== 8) {
                fout->yco.name = fin->yco.name + "_min";
                fout->yco.unit = fin->yco.unit;
            } else if (mod== 9) {
                fout->yco.name = "sum " + fin->yco.name;
                fout->yco.unit = fin->yco.unit;
            } else if (mod==10) {
                fout->yco.name = 
                    "I d" + fin->xco.name + " " + fin->yco.name;
                if( fin->yco.unit==fin->xco.unit+"-1" ||
                    fin->xco.unit==fin->yco.unit+"-1" )
                    fout->yco.unit = "";
                else
                    fout->yco.unit = fin->yco.unit + "*" + fin->xco.unit;
            } else if (mod==11) {
                fout->yco.name = fin->xco.name + 
                    string("[integral{") + 
                    fin->yco.name + "} < " + S(arg2) + "]";
                fout->yco.unit = fin->xco.unit;
            } else if (mod==13) {
                fout->yco.name = "max " + fin->xco.name +
                    " with positive " + fin->yco.name;
                fout->yco.unit = fin->xco.unit;
            } else if (mod==15) {
                fout->yco.name = fin->xco.name + 
                    "[center of gravity of " + 
                    fin->yco.name + "]";
                fout->yco.unit = fin->xco.unit;
            } else if (mod==16) {
                fout->yco.name = "sigma(" + fin->xco.name + ")";
                fout->yco.unit = "" ;
            } else if (mod==17) {
                fout->yco.name = "rho(" + fin->xco.name +
                    "|" + fin->yco.name + ")";
                fout->yco.unit = "" ;
            } else {
                throw "invalid mode";
            }
            fout->lDoc.push_back( "oio "+S(mod)+" # y = "+fout->yco.name );
        }

        PSpec eout( new CSpec );
        eout->z = fin->V[0]->z;
        eout->z.pop_back();
        for (uint j=0; j<fin->nJ(); j++) {
            double xval, zval, result;
            result = NIntOld::evaluate( fin->VS(j) );
            if (!savable) {
                cout << "results: " << result << "\n";
                continue;
            }
            xval = fin->V[j]->z[nz-1];
            eout->push_xy(xval, result);
            if (nz>=2) { // new spectrum if jump in other z values
                zval = fin->V[j]->z[nz-2];
                if ((j+1)<fin->nJ() && 
                    fin->V[j+1]->z[nz-2]!=zval) {
                    fout->V.push_back(eout);
                    eout = PSpec( new CSpec );
                    eout->z = fin->V[j+1]->z;
                    eout->z.pop_back();
                }
            }
        }
        if (savable) {
            fout->V.push_back(eout);
            NOlm::mem_store( fout );
        }
    }
};

double NIntOld::evaluate( const PSpec& ein )
{
    int i, i0;
    double r0, r1, r2, s0, s1, v, dr, area, triangle, p, center;
    int n = ein->size();
    switch(mod) {
    case  2:
        if (iarg2<0 || iarg2>=n) 
            throw "i out of range";
        return ein->y[iarg2];
    case  3:
        r0 = 0;
        for (i=0; i<n; i++) {
            r0 += ein->y[i];
        }
        return r0/n;
    case 4:
        r0 = 0; s0 = 0;
        for (i=0; i<n; i++) {
            r0 += ein->y[i] / n;
        }
        for (i=0; i<n; i++) {
            s0 += SQR(ein->y[i]-r0) / n;
        }
        return sqrt(s0);
    case  5:
        i0 = 0;
        for (i=1; i<n; i++) {
            if( ein->y[i]>ein->y[i0] ) i0 = i;
        }
        return ein->x[i0];
    case  6:
        i0 = 0;
        for (i=1; i<n; i++) {
            if( ein->y[i]>ein->y[i0] ) i0 = i;
        }
        return ein->y[i0];
    case  7:
        i0 = 0;
        for (i=1; i<n; i++) {
            if( ein->y[i]<ein->y[i0] ) i0 = i;
        }
        return ein->x[i0];
    case  8:
        i0 = 0;
        for (i=1; i<n; i++) {
            if( ein->y[i]<ein->y[i0] ) i0 = i;
        }
        return ein->y[i0];
    case  9: // sum_i y
        r0 = 0;
        for (i=0; i<n; i++)
            r0 += ein->y[i];
        return r0;
    case 10: // integral dx y
        r0 = 0;
        for (i=0; i<n-1; i++) {
            if (ein->x[i+1]<=ein->x[i])
                throw "x not sorted";
            r0 += (ein->x[i+1]-ein->x[i])*(ein->y[i+1]+ein->y[i])/2;
        }
        return r0;
    case 11:
        r0 = 0; r1 = 0;
        for (i=0; i<n-1; i++) {
            if (ein->x[i+1]<=ein->x[i])
                throw "x not sorted";
            r0 += (ein->x[i+1]-ein->x[i])*(ein->y[i+1]+ein->y[i])/2;
        }
        if (r0<=0)
            throw "int <= 0";
        for (i=0; i<n-1; i++) {
            dr = (ein->x[i+1]-ein->x[i])*(ein->y[i+1]+ein->y[i])/2;
            if ((r1+dr)/r0>=arg2) break;
            r1 += dr;
        }
        area = arg2*r0 - r1;
        triangle = (ein->y[i+1]-ein->y[i])/(ein->x[i+1]-ein->x[i])/2;
        if (triangle==0) {
            return ein->x[i]+area/dr;
        } else {
            p = ein->y[i]/2/triangle;
            return ein->x[i] - p + 
                (triangle>0?+1:-1) * sqrt(p*p + area/triangle);
        }
    case 13:
        for (i=n-1; i>=0; i--)
            if (ein->y[i]>0) break;
        return ein->x[i];
    case 15: // center of gravity
        r0 = 0; r1 = 0;
        for (i=0; i<n; i++) {
            r0 += ein->y[i];
            r1 += ein->y[i] * ein->x[i];
        }
        return r0>0 ? r1/r0 : 0;
    case 16: // stddev in x
        r0 = 0; r1 = 0; r2 = 0;
        for (i=0; i<n; i++) {
            r0 += ein->y[i];
            r1 += ein->y[i] * ein->x[i];
        }
        if( r0<=0 )
            throw "r0 <= 0";
        center = r1/r0;
        for (i=0; i<n; i++) {
            r2 += ein->y[i] * SQR(ein->x[i]-center);
        }
        return sqrt( r2/r0 );
    case 17: // correlation coefficient
        r0 = 0; r1 = 0; s0 = 0; s1 = 0; v = 0;
        for (i=0; i<n; i++) {
            r0 += ein->x[i] / n;
            r1 += ein->y[i] / n;
        }
        for (i=0; i<n; i++) {
            s0 += SQR(ein->x[i]-r0) / n;
            s1 += SQR(ein->y[i]-r1) / n;
            v  += (ein->x[i]-r0) * (ein->y[i]-r1) / n;
        }
        return v / sqrt(s0*s1);
    default:
        throw "BUG: this oio mode not implemented"; 
    }
}


//***************************************************************************//
//* functional operations                                                   *//
//***************************************************************************//

//! Functional integration, derivation, ...

void NOperate::Functional( const string& subcmd )
{
    NOlm::IterateD fiter;
    while ( POld fin=fiter() ) {
        POld fout( fin->new_old() );
        fout->lDoc.push_back( "of"+subcmd );

        if        ( subcmd=="i" ) { // integrate
            fout->yco.name = "Int d" + fin->xco.name + " " + fin->yco.name;
            fout->yco.unit = fin->xco.unit + " * " + fin->yco.unit;
        } else if ( subcmd=="d" ) { // derive
            fout->yco.name = "d" + fin->yco.name + " / d" + fin->xco.name;
            fout->yco.unit = fin->yco.unit + " / " + fin->xco.unit;
        } else if ( subcmd=="ac" ) { // autocorrelate
            fout->yco.name = "ac{" + fin->yco.name + "}";
            fout->yco.unit = fin->yco.unit + "^2";
        } else {
            throw "invalid 'of' command";
        }

        for ( uint j=0; j<fin->nJ(); j++ ) {
            PSpec sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->copy_z_base( sin );
            uint n = sin->size();

            if        ( subcmd=="i" ) { // integrate
                sout->x = sin->x;
                sout->y.resize( n );
                sout->y[0] = 0;
                for ( uint i=1; i<n; ++i ) {
                    sout->y[i] = sout->y[i-1] + 
                        (sin->x[i] - sin->x[i-1]) * 
                        (sin->y[i] + sin->y[i-1]) / 2;
                }
            } else if ( subcmd=="d" ) { // derive
                for( uint i=0; i<n-1; ++i ){
                    double dx = sin->x[i+1] - sin->x[i];
                    sout->push_xy(
                        (sin->x[i+1] + sin->x[i] ) / 2,
                        dx==0 ? 0 : (sin->y[i+1] - sin->y[i]) / dx );
                }
            } else if ( subcmd=="ac" ) { // autocorrelate
                throw "currently disabled, old code in"
                    " old/src-fragm/opr-func-ac.cpp";
            }
            fout->V.push_back( sout );
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}
